# Copyright Philipp Jund (jundp@cs.uni-freiburg.de) 2017. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
#
# These tests mostly follow moble's quaternion tests (under MIT license):
# https://github.com/moble/quaternion/blob/master/test/test_quaternion.py

from __future__ import division
import itertools
import numpy as np
import tensorflow as tf
import tfquaternion as tfq
from auto_eval_testcase import AutoEvalTestCase


# ____________________________________________________________________________
#                    functions to create testing quaternions
def get_quaternions():
    """ Returns two np.arrays of testing quaternions as np.array and
        Quaternion.
    """
    q_nan1 = np.array([np.nan, 0., 0., 0.], dtype=np.float32)
    q_inf1 = np.array([np.inf, 0., 0., 0.], dtype=np.float32)
    q_minf1 = np.array([-np.inf, 0., 0., 0.], dtype=np.float32)
    q_0 = np.array([0., 0., 0., 0.], dtype=np.float32)
    q_1 = np.array([1., 0., 0., 0.], dtype=np.float32)
    x = np.array([0., 1., 0., 0.], dtype=np.float32)
    y = np.array([0., 0., 1., 0.], dtype=np.float32)
    z = np.array([0., 0., 0., 1.], dtype=np.float32)
    q = np.array([1.1, 2.2, 3.3, 4.4], dtype=np.float32)
    qneg = np.array([-1.1, -2.2, -3.3, -4.4], dtype=np.float32)
    qbar = np.array([1.1, -2.2, -3.3, -4.4], dtype=np.float32)
    qnormalized = np.array([0.18257418583505537115232326093360,
                            0.36514837167011074230464652186720,
                            0.54772255750516611345696978280080,
                            0.73029674334022148460929304373440],
                            dtype=np.float32)
    qlog = np.array([1.7959088706354, 0.515190292664085,
                     0.772785438996128, 1.03038058532817], dtype=np.float32)
    qexp = np.array([2.81211398529184, -0.392521193481878,
                     -0.588781790222817, -0.785042386963756], dtype=np.float32)
    qmultidim = np.array([[1.1,  2.2, 3.3, 4.4], [1.1,  2.2, 3.3, 4.4]],
                         dtype=np.float32)
    np_quats = np.array([q_nan1, q_inf1, q_minf1, q_0, q_1, x, y, z, q, qneg,
                         qbar, qnormalized, qlog, qexp, qmultidim],
                         dtype=object)
    tf_quats = np.array([tfq.Quaternion(q_np) for q_np in np_quats])
    return np_quats, tf_quats


_qs, _ = get_quaternions()
N = len(_qs)
# lists of indices
(Q_NAN1, Q_INF1, Q_MINF1, Q_0, Q_1, X, Y, Z, Q,
 QNEG, QBAR, QNORMALIZED, QLOG, QEXP, QMULTIDIM) = range(N)
QS_ZERO = [i for i in range(N) if not np.flatnonzero(_qs[i]).any()]
QS_NONZERO = [i for i in range(N) if np.flatnonzero(_qs[i]).any()]
QS_NAN = [i for i in range(N) if np.isnan(_qs[i]).any()]
QS_NONNAN = [i for i in range(N) if not np.isnan(_qs[i]).any()]
QS_NONNANNONZERO = [i for i in range(N) if (not np.isnan(_qs[i]).any()
                                            and np.flatnonzero(_qs[i]).any())]
QS_INF = [i for i in range(N) if np.isinf(_qs[i]).any()]
QS_NONINF = [i for i in range(N) if not np.isinf(_qs[i]).any()]
QS_NONINFNONZERO = [i for i in range(N) if (not np.isinf(_qs[i]).any()
                                            and np.flatnonzero(_qs[i]).any())]
QS_FINITE = [i for i in range(N) if np.isfinite(_qs[i]).all()]
QS_FINITENONMULTI = [i for i in range(N) if np.isfinite(_qs[i]).all()
                                            and _qs[i].shape == (4,)]
QS_NONFINITE = [i for i in range(N) if not np.isfinite(_qs[i]).all()]
QS_FINITENONZERO = [i for i in range(N) if (np.isfinite(_qs[i]).all()
                                            and np.flatnonzero(_qs[i]).any())]

# ____________________________________________________________________________
#
class TfquaternionTest(AutoEvalTestCase):
    """ Tests for the module functions """

    def test_vector3d_to_quaternion(self):
         vec = np.array([1, 2, 3], dtype=np.float32)
         result =  np.array([0, 1, 2, 3], dtype=np.float32)
         quat = tfq.vector3d_to_quaternion(vec)
         with self.test_session():
            self.assertAllEqual(quat, result)
            # repeat stacks to obtain multidimensional quaternions
            for i in range(3):
                vec = np.stack([vec, vec])
                quat = tfq.vector3d_to_quaternion(vec)
                result = np.stack([result, result])
                self.assertAllEqual(quat, result)

    def test_quaternion_multiply(self):
        qs_np, qs_tf = get_quaternions()
        with self.test_session() as sess:
            for q in qs_tf[QS_FINITE]:
                self.assertAllEqual(q * qs_tf[Q_1], q)
                self.assertAllEqual(q * 1.0, q)
                self.assertAllEqual(1.0 * q, q)
                self.assertAllEqual(0.0 * q, np.zeros(q.shape))
                self.assertAllEqual(0.0 * q, q * 0.0)
            # Check multiplication with scalar
            for s in [-3., -2.3, -1.2, -1., 0., 1.0, 1.2, 2.3, 3.]:
                for q in qs_tf[QS_FINITE]:
                    q_w, q_x,  q_y,  q_z = tf.unstack(q, axis=-1)
                    s_times_q = [s * q_w, s * q_x, s * q_y, s * q_z]
                    s_times_q = tfq.Quaternion(tf.stack(s_times_q, axis=-1))
                    self.assertAllEqual(q * s, s_times_q)
                    self.assertAllEqual(s * q, q * s)
            # Check linearity (use placeholders to speed this up)
            pl1, pl2, pl3 = [tf.placeholder(tf.float32, (None, 4)) for i in range(3)]
            q1 , q2, q3 = [tfq.Quaternion(pl) for pl in [pl1, pl2, pl3]]
            ops = [q1 * (q2 + q3), (q1 * q2) + (q1 * q3),  # check 1
                   (q1 + q2) * q3, (q1 * q3) + (q2 * q3)]  # check 2
            triplets = itertools.permutations(qs_np[QS_FINITENONMULTI], 3)
            for q1np, q2np, q3np in triplets:
                q1np = q1np.reshape(1, 4) if q1np.shape == (4,) else q1np
                q2np = q2np.reshape(1, 4) if q2np.shape == (4,) else q2np
                q3np = q3np.reshape(1, 4) if q3np.shape == (4,) else q3np
                fd = { pl1: q1np, pl2: q2np, pl3: q3np }
                result = sess.run(ops, feed_dict=fd)
                # checks q1 * (q2 + q3) == (q1 * q2) + (q1 * q3),
                self.assertAllClose(result[0], result[1])
                # checks (q1 + q2) * q3 == (q1 * q3) + (q2 * q3)
                self.assertAllClose(result[2], result[3])
            # Check the multiplication table
            for q in [qs_tf[Q_1], qs_tf[X], qs_tf[Y], qs_tf[Z]]:
                self.assertAllEqual(qs_tf[Q_1] * q, q)
                self.assertAllEqual(q * qs_tf[Q_1], q)
            self.assertAllEqual(qs_tf[X] * qs_tf[X], -qs_tf[Q_1])
            self.assertAllEqual(qs_tf[X] * qs_tf[Y], qs_tf[Z])
            self.assertAllEqual(qs_tf[X] * qs_tf[Z], -qs_tf[Y])
            self.assertAllEqual(qs_tf[Y] * qs_tf[X], -qs_tf[Z])
            self.assertAllEqual(qs_tf[Y] * qs_tf[Y], -qs_tf[Q_1])
            self.assertAllEqual(qs_tf[Y] * qs_tf[Z], qs_tf[X])
            self.assertAllEqual(qs_tf[Z] * qs_tf[X], qs_tf[Y])
            self.assertAllEqual(qs_tf[Z] * qs_tf[Y], -qs_tf[X])
            self.assertAllEqual(qs_tf[Z] * qs_tf[Z], -qs_tf[Q_1])
            self.assertAllEqual(qs_tf[Z] * qs_tf[Z], -qs_tf[Q_1])

    def test_quaternion_divide(self):
        qs_np, qs_tf = get_quaternions()
        with self.test_session() as sess:
            # Check scalar division
            for q in qs_tf[QS_FINITENONZERO]:
                # use + np.zeros to broadcast qs_np[Q_1] to the multidim shape
                self.assertAllClose(q / q, qs_np[Q_1] + np.zeros(q.shape))
                self.assertRaises(TypeError, tfq.quaternion_divide, 1, q)
                self.assertAllClose(1.0 / q, q.inverse())
                self.assertAllClose([1.0] / q, q.inverse())
                self.assertAllClose(0.0 / q, qs_np[Q_0] + np.zeros(q.shape))
                for s in [-3., -2.3, -1.2, -1., 0., 1.0, 1.2, 2.3, 3.]:
                    self.assertAllClose(s / q, s * (q.inverse()))
            for q in qs_tf[QS_NONNAN]:
                self.assertAllClose(q / 1.0, q)
                for s in [-3., -2.3, -1.2, -1., 0., 1.0, 1.2, 2.3, 3.]:
                    # use np.array(1) to allow division by zero
                    gt = q * (np.array(1.0) / s).astype(np.float32)
                    self.assertAllClose(q / s, gt)

            # Check linearity
            pl1, pl2, pl3 = [tf.placeholder(tf.float32, (None, 4)) for i in range(3)]
            q1 , q2, q3 = [tfq.Quaternion(pl) for pl in [pl1, pl2, pl3]]
            ops = [(q1 + q2) / q3, (q1 / q3) + (q2 / q3)]
            for q1np, q2np in itertools.permutations(qs_np[QS_FINITE], 2):
                for q3np in qs_np[QS_FINITENONZERO]:
                    q1np = q1np.reshape(1, 4) if q1np.shape == (4,) else q1np
                    q2np = q2np.reshape(1, 4) if q2np.shape == (4,) else q2np
                    q3np = q3np.reshape(1, 4) if q3np.shape == (4,) else q3np
                    fd = {pl1: q1np, pl2: q2np, pl3: q3np}
                    a, b = sess.run(ops, feed_dict=fd)
                    # checks (q1 + q2) / q3 == (q1 / q3) + (q2 / q3)
                    self.assertAllClose(a, b)

            # Check the multiplication table
            for q in [qs_tf[Q_1], qs_tf[X], qs_tf[Y], qs_tf[Z]]:
                self.assertAllClose(qs_tf[Q_1] / q, q.conj())
                self.assertAllClose(q / qs_tf[Q_1], q)
            self.assertAllClose(qs_tf[X] / qs_tf[X], qs_tf[Q_1])
            self.assertAllClose(qs_tf[X] / qs_tf[Y], -qs_tf[Z])
            self.assertAllClose(qs_tf[X] / qs_tf[Z], qs_tf[Y])
            self.assertAllClose(qs_tf[Y] / qs_tf[X], qs_tf[Z])
            self.assertAllClose(qs_tf[Y] / qs_tf[Y], qs_tf[Q_1])
            self.assertAllClose(qs_tf[Y] / qs_tf[Z], -qs_tf[X])
            self.assertAllClose(qs_tf[Z] / qs_tf[X], -qs_tf[Y])
            self.assertAllClose(qs_tf[Z] / qs_tf[Y], qs_tf[X])
            self.assertAllClose(qs_tf[Z] / qs_tf[Z], qs_tf[Q_1])

    def test_quaternion_conjugate(self):
        qs_np, qs_tf = get_quaternions()
        with self.test_session():
            self.assertAllEqual(qs_tf[Q].conjugate(), qs_tf[QBAR])
            for q in qs_tf[QS_NONNAN]:
                self.assertAllEqual(q.conjugate(), q.conj())
                self.assertAllEqual(q.conjugate().conjugate(), q)
                cw, cx, cy, cz = tf.unstack(q.conjugate(), axis=-1)
                qw, qx, qy, qz = tf.unstack(q, axis=-1)
                self.assertAllEqual(cw, qw)
                self.assertAllEqual(cx, -qx)
                self.assertAllEqual(cy, -qy)
                self.assertAllEqual(cz, -qz)

    def test_rotate_vector_by_quaternion(self):
        v = [3., 4., 5.]
        qs_np, qs_tf = get_quaternions()
        with self.test_session():
            for q in qs_tf[QS_FINITENONZERO]:
                rotated = q * tfq.vector3d_to_quaternion(v) * q.inverse()
                rotated = tfq.quaternion_to_vector3d(rotated)
                self.assertAllClose(tfq.rotate_vector_by_quaternion(q, v),
                                    rotated)

    def test_quaternion_to_tensor(self):
        ref = tfq.Quaternion([1.0, 0.0, 0.0, 0.0])
        with self.test_session():
            self.assertAllEqual(type(tf.convert_to_tensor(ref)), tf.Tensor)
            # reduce sum internally calls tf.convert_to_tensor
            self.assertAllEqual(tf.reduce_sum(ref), 1.0)

class QuaternionTest(AutoEvalTestCase):
    """ Tests for the member functions of class Quaternion """

    def test___init__(self):
        ref = tfq.Quaternion([1.0, 0.0, 0.0, 0.0])
        val = (1.0, 0.0, 0.0, 0.0)
        variable = tf.Variable(val)
        with self.test_session() as sess:
            sess.run(tf.global_variables_initializer())
            #default constructor
            self.assertAllEqual(tfq.Quaternion(), ref)
            # from variable
            self.assertAllEqual(tfq.Quaternion(variable), ref)
            # from constant
            self.assertAllEqual(tfq.Quaternion(tf.constant(val)), ref)
            # from np.array
            self.assertAllEqual(tfq.Quaternion(np.array(val)), ref)
            # from list
            self.assertAllEqual(tfq.Quaternion([1.0, 0.0, 0.0, 0.0]), ref)
            self.assertAllEqual(tfq.Quaternion([1, 0, 0, 0]), ref)
            # from tuple
            self.assertAllEqual(tfq.Quaternion(val), ref)
            # from Quaternion
            self.assertAllEqual(tfq.Quaternion(ref), ref)
            # wrong shape
            self.assertRaises(ValueError, tfq.Quaternion, [[1, 0]])
            # wrong type
            self.assertRaises(TypeError, tfq.Quaternion, val, dtype=tf.int32)
            self.assertRaises(TypeError, tfq.Quaternion, [1, 0, 0, 0],
                              dtype=tf.int32)

    def test_value(self):
        val = [1.0, 2.0, 3.0, 4.0]
        q = tfq.Quaternion(val)
        with self.test_session():
            self.assertAllEqual(q, tf.constant(val))


    def test___add__(self):
        a = tfq.Quaternion([1, 2, 3, 4])
        b = tfq.Quaternion([5, 6, 7, 8])
        result = np.array([6, 8, 10, 12], dtype=np.float32)
        with self.test_session():
            self.assertAllEqual(a + b, result)

    def test___sub__(self):
        a = tfq.Quaternion([1, 2, 3, 4])
        b = tfq.Quaternion([5, 6, 7, 8])
        result = np.array([-4, -4, -4, -4], dtype=np.float32)
        with self.test_session():
            self.assertAllEqual(a - b, result)

    def test___rmul__(self):
        qs_np, qs_tf = get_quaternions()
        # Note that quaternion multiplication is only applied if both
        # multiplicands are quaternions. Also note that multiplying
        # by a np.array returns one Quaternion for each entry in the np.array,
        # therefore we don't test this here
        with self.test_session():
            self.assertAllEqual(qs_tf[X] * qs_tf[X], -qs_tf[Q_1])
            self.assertAllEqual(qs_tf[X] * 2.0, [0.0, 2.0, 0.0, 0.0])
            self.assertAllClose(qs_np[X] * qs_tf[X], [0.0, 1.0, 0.0, 0.0])
            self.assertAllClose(qs_np[Q] * qs_tf[Q], qs_np[Q] * qs_np[Q])

    def test___rdiv__(self):
        qs_np, qs_tf = get_quaternions()
        with self.test_session():
            self.assertAllClose(qs_tf[X] / qs_tf[X], qs_tf[Q_1])
            self.assertAllClose(qs_tf[X] / 2.0, [0.0, 0.5, 0.0, 0.0])

    def test___lt__(self):
        ref = tf.constant(np.arange(4), dtype=tf.float32)
        ref2 = tf.constant(np.arange(2, 6), dtype=tf.float32)
        ref3 = tf.constant(np.arange(-2, 2), dtype=tf.float32)
        ref_q = tfq.Quaternion(ref)
        with self.test_session():
            self.assertAllFalse(ref < ref)
            self.assertAllTrue(ref < ref2)
            self.assertAllTrue(ref3 < ref)

    def test___le__(self):
        ref = tf.constant(np.arange(4), dtype=tf.float32)
        ref2 = tf.constant(np.arange(2, 6), dtype=tf.float32)
        ref3 = tf.constant(np.arange(-2, 2), dtype=tf.float32)
        ref_q = tfq.Quaternion(ref)
        with self.test_session():
            self.assertAllTrue(ref <= ref)
            self.assertAllTrue(ref <= ref2)
            self.assertAllTrue(ref3 <= ref)

    def test___eq__(self):
        ref = tf.constant(np.arange(4), dtype=tf.float32)
        ref2 = tf.constant(np.arange(2, 6), dtype=tf.float32)
        ref_q = tfq.Quaternion(ref)
        with self.test_session():
            self.assertAllTrue(ref_q == ref)
            self.assertAllTrue(ref_q == ref_q)
            self.assertAllTrue(ref_q == tfq.Quaternion(ref))
            self.assertAllFalse(ref_q == ref2)

    def test___ne__(self):
        ref = tf.constant(np.arange(4), dtype=tf.float32)
        ref2 = tf.constant(np.arange(2, 6), dtype=tf.float32)
        ref_q = tfq.Quaternion(ref)
        with self.test_session():
            self.assertAllFalse(ref_q != ref)
            self.assertAllFalse(ref_q != ref_q)
            self.assertAllFalse(ref_q != tfq.Quaternion(ref))
            self.assertAllTrue(ref_q != ref2)

    def test___gt__(self):
        ref = tf.constant(np.arange(4), dtype=tf.float32)
        ref2 = tf.constant(np.arange(2, 6), dtype=tf.float32)
        ref3 = tf.constant(np.arange(-2, 2), dtype=tf.float32)
        ref_q = tfq.Quaternion(ref)
        with self.test_session():
            self.assertAllFalse(ref > ref)
            self.assertAllFalse(ref > ref2)
            self.assertAllTrue(ref > ref3)

    def test___ge__(self):
        ref = tf.constant(np.arange(4), dtype=tf.float32)
        ref2 = tf.constant(np.arange(2, 6), dtype=tf.float32)
        ref3 = tf.constant(np.arange(-2, 2), dtype=tf.float32)
        ref_q = tfq.Quaternion(ref)
        with self.test_session():
            self.assertAllTrue(ref >= ref)
            self.assertAllFalse(ref >= ref2)
            self.assertAllTrue(ref >= ref3)

    def test_inverse(self):
        qs_np, qs_tf = get_quaternions()
        with self.test_session() as sess:
            for q in qs_tf[QS_FINITENONZERO]:
                gt = qs_np[Q_1] + np.zeros(q.shape)
                self.assertAllClose(q * q.inverse(), gt)

    def test_normalized(self):
        # raise NotImplementedError("Test not implemented")
        pass

    def test_as_rotation_matrix(self):
        # raise NotImplementedError("Test not implemented")
        pass

    def test__validate_shape(self):
        self.assertRaises(ValueError, tfq.Quaternion, [1,2,3])

    def test__validate_type(self):
        self.assertRaises(TypeError, tfq.Quaternion, dtype=tf.int32)

    def test_norm(self):
        with self.test_session():
            self.assertAllEqual(tfq.Quaternion((1, 2, 3, 4)).norm(), [30.0])
            self.assertAllEqual(tfq.Quaternion((-1, -2, -3, -4)).norm(), [30.0])
            self.assertAllEqual(tfq.Quaternion((0, 0, 0, 0)).norm(), [0.0])

    def test_abs(self):
        delta = 0.00001
        with self.test_session(use_gpu=True):
            self.assertAlmostEqual(tfq.Quaternion((1, 2, 3, 4)).abs(),
                                   5.47722, delta=delta)
            self.assertAlmostEqual(tfq.Quaternion((-1, -2, -3, -4)).abs(),
                                   5.47722, delta=delta)
            self.assertAllEqual(tfq.Quaternion((0, 0, 0, 0)).abs(), 0.0)


if __name__ == "__main__":
    tf.test.main()



# Copyright Philipp Jund (jundp@cs.uni-freiburg.de) 2017. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""

This small library implements quaternion operations in tensorflow.
All operations are differentiable.

"""
import tensorflow as tf


# ____________________________________________________________________________
#                     Quaternion module functions
def scope_wrapper(func, *args, **kwargs):
    """ Creates a tf name scope around the function with its name. """
    def scoped_func(*args, **kwargs):
        with tf.name_scope("quaternion_{}".format(func.__name__)):
            return func(*args, **kwargs)
    return scoped_func


@scope_wrapper
def vector3d_to_quaternion(x):
    """ Converts a tensor of 3D vectors to a quaternion by prepending a 0
        to the last dimension.

    Args:
        x: A `tf.Tensor` of rank R, the last dimension must be 3.

    Returns:
        A `Quaternion` of Rank R with the last dimension being 4.

    Raises:
        ValueError, if the last dimension of x is not 3.
    """
    x = tf.convert_to_tensor(x)
    if x.shape[-1] != 3:
        raise ValueError("The last dimension of x must be 3.")
    Quaternion.validate_type(x)
    return Quaternion(tf.pad(x, (len(x.shape) - 1) * [[0, 0]] + [[1, 0]]))


@scope_wrapper
def quaternion_to_vector3d(q):
    """ Removes the w component(s) of quaternion(s) q. """
    return q.value()[..., 1:]


@scope_wrapper
def _prepare_tensor_for_div_mul(x):
    """ Prepares the tensor x for division/multiplication by
        a) converting a to a tensor if necessary
        b) prepending a 0 in the last dimension if the last dimension is 3
        c) validating the type and shape
    """
    x = tf.convert_to_tensor(x)
    if x.shape[-1] == 3:
        x = vector3d_to_quaternion(x)
    Quaternion.validate_shape(x)
    Quaternion.validate_type(x)
    return x


@scope_wrapper
def quaternion_multiply(a, b):
    """ Multiplies two quaternion tensors. Note that this differs from
        tf.multiply and is not commutative.

    Args:
        a, b: A `tf.Tensor` with shape (..., 4).

    Returns:
        A `Quaternion`.
    """
    a = _prepare_tensor_for_div_mul(a)
    b = _prepare_tensor_for_div_mul(b)
    w1, x1, y1, z1 = tf.unstack(a, axis=-1)
    w2, x2, y2, z2 = tf.unstack(b, axis=-1)
    w = w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2
    x = w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2
    y = w1 * y2 + y1 * w2 + z1 * x2 - x1 * z2
    z = w1 * z2 + z1 * w2 + x1 * y2 - y1 * x2
    return Quaternion(tf.squeeze(tf.stack((w, x, y, z), axis=-1)))


@scope_wrapper
def quaternion_divide(a, b):
    """ Divides a by quaternion tensor b. A may be a scalar value

    Args:
        a, b: A `tf.Tensor` with shape (..., 4).

    Returns:
        A `Quaternion`.
    """
    a = tf.convert_to_tensor(a)
    if a.shape == () or a.shape[-1] == 1:  # scalar
        return Quaternion(tf.multiply(a, b.conj()) / Quaternion(b).norm())
    bnorm = tf.squeeze(Quaternion(b).norm())
    w1, x1, y1, z1 = tf.unstack(a, axis=-1)
    w2, x2, y2, z2 = tf.unstack(b, axis=-1)
    w = (w1*w2 + x1*x2 + y1*y2 + z1*z2) / bnorm
    x = (-w1*x2 + x1*w2 - y1*z2 + z1*y2) / bnorm
    y = (-w1*y2 + x1*z2 + y1*w2 - z1*x2) / bnorm
    z = (-w1*z2 - x1*y2 + y1*x2 + z1*w2) / bnorm
    return Quaternion(tf.squeeze(tf.stack((w, x, y, z), axis=-1)))


@scope_wrapper
def quaternion_conjugate(q):
    """ Computes the conjugate of q, i.e. returns [w, -x, -y, -z]. """
    return Quaternion(tf.multiply(q, [1.0, -1.0, -1.0, -1.0]))


@scope_wrapper
def rotate_vector_by_quaternion(q, v):
    """ Rotates a vector (or tensor with last dimension of 3) by q,
        i.e. it computes v' = q * v * conjugate(q). 
        Faster version, found here:
        https://blog.molecular-matters.com/2013/05/24/a-faster-quaternion-vector-multiplication/
    
    Args:
        q: A `Quaternion` or `tf.Tensor` with shape (..., 4)
        v: A `tf.Tensor` with shape (..., 3)

    """
    return quaternion_to_vector3d(q * vector3d_to_quaternion(v) * q.inverse())
    #v = tf.convert_to_tensor(v)
    #q = q.normalized()
    #w = q.value()[..., 0]
    #print("w.shape", w.shape)
    #print("v.shape", w.shape)
    #q_xyz = q.value()[..., 1:]
    #print("q_xyz.shape", q_xyz.shape)
    #t = 2 * tf.cross(q_xyz, v)
    #print("t.shape", t.shape)
    #return v + tf.expand_dims(w, axis=-1) * t + tf.cross(q_xyz, t)


# ____________________________________________________________________________
#                      The quaternion class
class Quaternion(object):
    """ Multidimensional quaternion. The API resembles that of tf.Variable. """
    
    # When trying to scale the components of the Quaternion individually
    # by right-multiplying a tf.Quaternion with a 4-dimensional np.array a, the
    # default numpy behaviour is to call `Quaternion.__rmul__(i)` for each
    # element i in a, resulting in 4 tfq.Quaternions instead of one.
    # Setting __array_priority__ = 1000 fixes this. (For further reference see
    # https://stackoverflow.com/questions/40694380/forcing-multiplication-to-use-rmul-instead-of-numpy-array-mul-or-byp)
    __array_priority__ = 1000

    def __init__(self, wxyz=(1, 0, 0, 0), dtype=tf.float32, name=None):
        """
        Args:
            wxyz: The values for w, x, y, z, a `tf.Tensor` with shape (..., 4).
                Note that quaternions only support floating point numbers.
                Defaults to (1.0, 0.0, 0.0, 0.0)
            dtype: The type used for the quaternion, must be a floating point
                number, i.e. one of tf.float16, tf.float32, tf.float64.
            name: An optional name for the tensor.

        Returns:
            A Quaternion.

        Raises:
            ValueError, if wxyz is a `tf.Tensor` and the tensors dtype differs
                from the given dtype.
            ValueError, if the last dimension of wxyz is not 4.
            TypeError, if dtype is not a float.
        """
        self._q = tf.convert_to_tensor(wxyz, dtype=dtype, name=name)
        self.name = name if name else ""
        self.validate_type(self._q)
        self.validate_shape(self._q)  # check that shape is (..., 4)

    def value(self):
        """ Returns a `Tensor` which holds the value of the quaternion. Note
            that this does not return a reference, so you can not alter the
            quaternion through this.
        """
        return self._q

    def eval(self, session=None):
        """ In a session, computes and returns the value of this quaternion. """
        return self._q.eval(session=session)

    def _ref(self):
        return self._q._ref()

    @property
    def dtype(self):
        """The `DType` of this quaternion. """
        return self._q.dtype

    @property
    def op(self):
        """The `Operation` of this quaternion. """
        return self._q.op

    @property
    def graph(self):
        """The `Graph` of this quaternion. """
        return self._q.graph

    @property
    def shape(self):
        """The `TensorShape` of the variable. Is always [..., 4].

        Returns:
          A `TensorShape`.
        """
        return self._q.get_shape()

    def get_shape(self):
        """Alias of Quaternion.shape."""
        return self.shape

    def _as_graph_element(self):
        """Conversion function for Graph.as_graph_element()."""
        return self._q

    def __add__(self, other):
        return Quaternion(tf.add(self._q, tf.convert_to_tensor(other)))

    def __radd__(self, other):
        return Quaternion(tf.add(tf.convert_to_tensor(other), self._q))

    def __sub__(self, other):
        return Quaternion(tf.subtract(self._q, tf.convert_to_tensor(other)))

    def __rsub__(self, other):
        return Quaternion(tf.subtract(tf.convert_to_tensor(other), self._q))

    def __mul__(self, other):
        if isinstance(other, Quaternion):
            return quaternion_multiply(self, other)
        return Quaternion(tf.multiply(self._q, tf.convert_to_tensor(other)))

    def __rmul__(self, other):
        # This is only called when __mul__ fails, so 'other' can not
        # be a Quaternion.
        return Quaternion(tf.multiply(self._q, tf.convert_to_tensor(other)))

    def __div__(self, other):
        if isinstance(other, Quaternion):
            return quaternion_divide(self, other)
        return tf.divide(self._q, tf.convert_to_tensor(other))
        

    def __rdiv__(self, other):
        if (isinstance(other, Quaternion) or
                tf.convert_to_tensor(other).shape == () or
                tf.convert_to_tensor(other).shape[-1] == 1):  # scalar
            return quaternion_divide(other, self)
        return tf.divide(tf.convert_to_tensor(other), self._q)

    def __truediv__(self, other):
        return self.__div__(other)

    def __rtruediv__(self, other):
        return self.__rdiv__(other)

    def __neg__(self):
        return Quaternion(-1 * self._q)

    # rich comparisons
    def __lt__(self, other):  # <
        return tf.less(self._q, other)

    def __le__(self, other):  # <=
        return tf.less_equal(self._q, other)

    def __eq__(self, other):  # ==
        return tf.equal(self._q, other)

    def __ne__(self, other):  # !=
        return tf.not_equal(self._q, other)

    def __gt__(self, other):  # >
        return tf.greater(self._q, other)

    def __ge__(self, other):  # >=
        return tf.greater_equal(self._q, other)

    def __repr__(self):
        return "<tfq.Quaternion '{}' ({})>".format(self.name,
                                                   self._q.__repr__()[1:-1])

    @scope_wrapper
    def conjugate(self):
        """ Computes the conjugate of self.q, i.e. returns [w, -x, -y, -z]. """
        return quaternion_conjugate(self)

    def conj(self):
        """ Computes the conjugate of self.q, i.e. returns [w, -x, -y, -z].
            Alias for Quaternion.conjugate().
        """
        return quaternion_conjugate(self)


    @scope_wrapper
    def inverse(self):
        return Quaternion(tf.convert_to_tensor(self.conjugate()) / self.norm())

    @scope_wrapper
    def normalized(self):
        return Quaternion(tf.divide(self._q, self.abs()))

    @scope_wrapper
    def as_rotation_matrix(self):
        """ Calculates the corresponding rotation matrix. See
            http://www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToMatrix/

        Returns:
            A `tf.Tensor` with R+1 dimensions and
            shape [d_1, ..., d_(R-1), 3, 3], the rotation matrix

        """
        # helper functions
        def diag(a, b):  # computes the diagonal entries,  1 - 2*a**2 - 2*b**2
            return 1 - 2 * tf.pow(a, 2) - 2 * tf.pow(b, 2)

        def tr_add(a, b, c, d):  # computes triangle entries with addition
            return 2 * a * b + 2 * c * d

        def tr_sub(a, b, c, d):  # computes triangle entries with subtraction
            return 2 * a * b - 2 * c * d

        w, x, y, z = tf.unstack(self.normalized().value(), axis=-1)
        m = [[diag(y, z), tr_sub(x, y, z, w), tr_add(x, z, y, w)],
             [tr_add(x, y, z, w), diag(x, z), tr_sub(y, z, x, w)],
             [tr_sub(x, z, y, w), tr_add(y, z, x, w), diag(x, y)]]
        return tf.stack([tf.stack(m[i], axis=-1) for i in range(3)], axis=-2)

    @staticmethod
    def validate_shape(x):
        """ Raises a value error if x.shape ist not (..., 4). """
        error_msg = ("Can't create a quaternion from a tensor with shape {}."
                     "The last dimension must be 4.")
        # Check is performed during graph construction. If your dimension
        # is unknown, tf.reshape(x, (-1, 4)) might work.
        if x.shape[-1] != 4:
            raise ValueError(error_msg.format(x.shape))

    @staticmethod
    def validate_type(x):
        """ Raises a type error if the dtype of x is not float. """
        if not x.dtype.is_floating:
            raise TypeError("Quaternion: dtype must be one of float16/32/64.")

    @scope_wrapper
    def norm(self):
        """ Returns the norm of the quaternion. """
        return tf.reduce_sum(tf.square(self._q), axis=-1, keep_dims=True)

    @scope_wrapper
    def abs(self):
        """ Returns the square root of the norm of the quaternion. """
        return tf.sqrt(tf.reduce_sum(tf.square(self._q)))


# ____________________________________________________________________________
#                          quaternion to tensor conversion
def quaternion_to_tensor(x, dtype=None, name=None, as_ref=None):
    """ Converts a Quaternion to a tensor. """
    # Todo(phil): handle as_ref correctly
    return tf.convert_to_tensor(x.value(), dtype, name)


tf.register_tensor_conversion_function(Quaternion, quaternion_to_tensor,
                                       priority=100)



# Copyright Philipp Jund (jundp@cs.uni-freiburg.de) 2017. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
#
#
#
# AutoEvalTestCase
# =======
# The AutoEvalTestCase automatically evaluates tensor-like objects
# such as tf.Variables, tf.Tensor, tf.SparseTensor etc. This is motivated by
# the fact that most tests are concerned with the values of the tensors and not
# the tf.Tensor object. Inheriting from this class removes the need for calling
# `.eval()` for each tensor, being easier on the eyes.
#
#
#  Example:
#  =======
#  class MyModuleTest(AutoEvalTestCase):
#
#      def test_myfunction():
#          gt = np.array([[1., 1.], [1., 1.]], dtype=tf.float32)
#          with self.test_session():
#              self.assertEqual(tf.ones((2, 2)), gt)
#              # you can still use the vanilla tf.test.TestCase when required
#              tf.test.TestCase().assertEqual(tf.ones((2, 2)).eval(), gt)
#


import tensorflow as tf


class AutoEvalTestCase(tf.test.TestCase):
    """ Unit test class that auto-evaluates tf.Tensors in assertions. """

    @staticmethod
    def _is_tensor(x):
        """ Returns true if x is tensor-like, including tfq.Quaternion. """
        return isinstance(x, (tf.Tensor, tf.SparseTensor, tf.Variable,
                              tfq.Quaternion))

    @staticmethod
    def _auto_eval(x):
        """ Evaluates x if it is a Tensor-like object otherwise returns x. """
        return x.eval() if AutoEvalTestCase._is_tensor(x) else x

    def assertEqual(self, a, b, msg=None):
        sup = super(AutoEvalTestCase, self)
        return sup.assertEqual(self._auto_eval(a), self._auto_eval(b), msg=msg)

    def assertAllEqual(self, a, b):
        sup = super(AutoEvalTestCase, self)
        return sup.assertAllEqual(self._auto_eval(a), self._auto_eval(b))

    def assertAlmostEqual(self, a, b, places=None, msg=None, delta=None):
        sup = super(AutoEvalTestCase, self)
        return sup.assertAlmostEqual(self._auto_eval(a), self._auto_eval(b),
                                     places=places, msg=msg, delta=delta)

    def assertAllClose(self, a, b, *args, **kwargs):
        sup = super(AutoEvalTestCase, self)
        return sup.assertAllClose(self._auto_eval(a), self._auto_eval(b),
                                  *args, **kwargs)

    def assertTrue(self, a, msg=None):
        sup = super(AutoEvalTestCase, self)
        return sup.assertTrue(self._auto_eval(a), msg)

    def assertFalse(self, a, msg=None):
        sup = super(AutoEvalTestCase, self)
        return sup.assertFalse(self._auto_eval(a), msg)

    def assertAllFalse(self, a):
        """ Checks for a given boolean np.array if all values are False. """
        a = self._GetNdArray(self._auto_eval(a))
        return self.assertFalse(a.any())

    def assertAllTrue(self, a):
        """ Checks for a given boolean np.array if all values are True. """
        a = self._GetNdArray(self._auto_eval(a))
        return self.assertTrue(a.all())



# Author: Philipp Jund (jundp@informatik.uni-freiburg.de)
from setuptools import setup, find_packages
import os

dirname = os.path.dirname(os.path.realpath(__file__))
with open(os.path.join(dirname, 'README.md')) as f:
    long_description = f.read()

setup(
    name='tfquaternion',
    version='0.1',
    description="A differentiable quaternion implementation in tensorflow.",
    long_description=long_description,
    url='https://github.com/PhilJd/tf-quaternion',

    author='Philipp Jund',
    author_email='jundp@cs.uni-freiburg.de',

    keywords='quaternion tensorflow differentiable',
    packages=find_packages(),
    install_requires=['numpy', 'tensorflow'],

    license='Apache 2.0',

    classifiers=[
        # Change later on:
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering',

        'License :: OSI Approved :: Apache Software License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        # Todo(phil): update this
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)
